#include 	<16F84.H>

#use 		delay(clock=10000000)
#use		fast_io(A)
#use 		rs232(baud=19200, xmit=PIN_A1, rcv=PIN_A0)
#use		I2C(master, sda=PIN_B4, scl=PIN_B5, fast)

#define		SERVER_IP1				192
#define		SERVER_IP2				168
#define		SERVER_IP3				2
#define		SERVER_IP4				3

#define		SERVER_HTTPPORT_15to8	0
#define		SERVER_HTTPPORT_7to0	80

#define 	END 					0xC0
#define		ESC 					0xDB
#define		ESC_END 				0xDC
#define		ESC_ESC 				0xDD

#define		PROTOCOL_TCP			6

#define		CTS_PIN					PIN_A2

#define		TCP_FLAG_FIN			1
#define		TCP_FLAG_SYN			2
#define		TCP_FLAG_ACK			16

#define		STATUS					3
#define		C						0

#define		CLIENT_IP1				0
#define		CLIENT_IP2				1
#define		CLIENT_IP3				2
#define		CLIENT_IP4				3
#define		CLIENT_PORT_15to8		4
#define		CLIENT_PORT_7to0		5
#define		SERVER_PORT_15to8		6
#define		SERVER_PORT_7to0		7
#define		CLIENT_SEQ_31to24		8
#define		CLIENT_SEQ_23to16		9
#define		CLIENT_SEQ_15to8		10
#define		CLIENT_SEQ_7to0			11
#define		CLIENT_ACK_31to24		12
#define		CLIENT_ACK_23to16		13
#define		CLIENT_ACK_15to8		14
#define		CLIENT_ACK_7to0			15

#define		INDEX_DOC_LOCATION		0x0004	// the second entry in the eeprom file table points to the index



int		currentByte;
long	checksumTallyA, checksumTallyB;



void receiveSLIP()
{
	output_low( CTS_PIN );	// CTS pin output is inverted by the RS232 driver IC
	currentByte = getC();
	output_high( CTS_PIN );
}



// encode SLIP escape codes
void sendSLIP( int transmitByte )		
{
	if ( transmitByte == END )
	{
		putC( ESC );
		putC( ESC_END );
	}
	else
	{
		if ( transmitByte == ESC )
		{
			putC( ESC );
			putC( ESC_ESC );
		}
		else
		{
			putC( transmitByte );
		}
	}
}



void checksumCalc()
{
	long checksumTotal;
	checksumTotal = checksumTallyA & 0xFF;
	checksumTotal += checksumTallyB >> 8;
	checksumTotal = ( checksumTotal >> 8 ) | ( checksumTotal << 8 );		// swap bytes
	checksumTotal += checksumTallyB & 0xFF;
	checksumTotal += checksumTallyA >> 8;

	checksumTotal = ~checksumTotal;
	sendSLIP( checksumTotal >> 8 );
	sendSLIP( checksumTotal );
}



void i2c_startFunc()
{
	i2c_start();
}



void i2c_stopFunc()
{
	i2c_stop();
}



void main()
{
	int		headerArray[16];
	int		byteCounter;

	long	serverIPidentification;
	int		serverInitialSEQ31to24, serverInitialSEQ23to16, serverInitialSEQ15to8, serverInitialSEQ7to0;

	int		TCPflags;
	long	counter;
	long	replyLength, packetLength, fileLength;
	int		TCPheaderLength;
	short	data, checksumNotDone, again, sendThisByte, oddLength, gotHeader, dataRemaining;

	int		startByte;
	long	endByte;
	int		fileStartHigh, fileStartLow;
	int		httpRequestChar1, httpRequestChar2, httpRequestChar3;
	long	requestedFileEntry;

	int		highByte;

	int		offset;

	short	rst;

	set_tris_a( 1 );

start:

	byteCounter = 0;
	counter = 0;
	packetLength = 20;
	data = false;
	gotHeader = false;
	rst=false;

	while ( counter != packetLength )
	{

		receiveSLIP();

		if ( currentByte == END )
		{
			goto start;
		}
		else
		{
			if ( currentByte == ESC )
			{
				receiveSLIP();

				if ( currentByte == ESC_END )
				{
					currentByte = END;
				}
				else
				{
				currentByte = ESC;
				}
			}
		}


		if ( data )
		{	
			currentByte-=48;
			
			if ( counter==offset+5 )
			{
				httpRequestChar1=currentByte;
			}

			if ( counter==offset+6 )
			{
				httpRequestChar2 = currentByte;
			}

			if ( counter==offset+7 )
			{
				httpRequestChar3 = currentByte;
			}

		}
		else
		{
			
			
			switch( counter )
			{

			case 2:	
				packetLength = currentByte;
				packetLength <<= 8;
				packetLength += 4;
				break;

			case 3:
				packetLength += currentByte;
				packetLength -= 4;
				break;
  
			case 9:
				// only process TCP packets													
				if ( currentByte == PROTOCOL_TCP )
				{
					break;
				}
				goto start;

			case 12:
			case 13:
			case 14:
			case 15:
			case 20:
			case 21:
			case 22:
			case 23:
			case 24:
			case 25:
			case 26:
			case 27:
			case 28:
			case 29:
			case 30:
			case 31:
				headerArray[ byteCounter ] = currentByte;
				byteCounter++;
				break;	
				
			case 32:
				// store TCP header length value in BYTES											
				TCPheaderLength = ( currentByte & 0xF0 ) >> 2;	
				break;

			case 33:
				TCPflags = currentByte;

				offset = TCPheaderLength + 20;

				if ( packetLength != offset )
				{
					data = TRUE;
				}
			}		


		}
		

		counter ++;
		
	}

	// sending IP and TCP headers only
	if ( !data )												
	{	

		// opening a new connection
		if ( bit_test( TCPflags, 1 ) )							
		{
			// set the ack flag
			bit_set( TCPflags, 4 ); 
			
			if (( headerArray[ SERVER_PORT_15to8 ] == SERVER_HTTPPORT_15to8 ) && ( headerArray[ SERVER_PORT_7to0 ] == SERVER_HTTPPORT_7to0 ))
			{
				// HTTP port connection


				// add 64000 to our initial Sequence Number
				#asm

				movf	0xFA, W	
				btfsc	STATUS, C	
				incfsz	0xFA, W	
				addwf	serverInitialSEQ15to8, F	

				clrw
				btfsc	STATUS, C
				movlw	1
				addwf	serverInitialSEQ23to16, F

				clrw
				btfsc	STATUS, C
				movlw	1
				addwf	serverInitialSEQ31to24, F

				#endasm

				headerArray[ CLIENT_ACK_31to24 ] = serverInitialSEQ31to24;
				headerArray[ CLIENT_ACK_23to16 ] = serverInitialSEQ23to16;
				headerArray[ CLIENT_ACK_15to8 ] = serverInitialSEQ15to8;
				headerArray[ CLIENT_ACK_7to0 ] = serverInitialSEQ7to0;
			}
			else
			{
				// send reset (RST) for connection to invalid port number
				headerArray[ CLIENT_ACK_31to24 ] = 0;
				headerArray[ CLIENT_ACK_23to16 ] = 0;
				headerArray[ CLIENT_ACK_15to8 ] = 0;
				headerArray[ CLIENT_ACK_7to0 ] = 0;

				bit_set( TCPflags, 2 );
				bit_clear( TCPflags, 1 );
			}

		}
		else
		{
			// ignore dataless ACKs
			if ( !bit_test( TCPflags, 0 ) )
			{
				goto start;										
			}	
			
		}
	}
	else
	{
		// what page is requested?
		if ( httpRequestChar1 == 0xF0 )
		{
			requestedFileEntry = 0x04;
		}
		else
		{
			if (( httpRequestChar1 > 9 ) || ( httpRequestChar2 > 9 ) || ( httpRequestChar3 != 0xF0 ))
			{
				requestedFileEntry = 0;
			}
			else
			{
				requestedFileEntry = httpRequestChar1;
				requestedFileEntry <<= 8;
				httpRequestChar1 <<= 1;
				requestedFileEntry += httpRequestChar1;
				requestedFileEntry += httpRequestChar2;
				requestedFileEntry <<= 2;
			}
		}


		// get file details from EEPROM fat
		i2c_startFunc();
		i2c_write( 0b10100000 );
		i2c_write( requestedFileEntry >>8 );
		i2c_write( requestedFileEntry ); 
		i2c_startFunc();
		i2c_write( 0b10100001 );

		fileStartHigh = i2c_read();
		fileStartLow = i2c_read();

		fileLength = i2c_read();

		fileLength <<= 8;
		fileLength += i2c_read( 0 );

		i2c_stopFunc();




	}

																						
	TCPheaderLength +=20;
	if ( bit_test( TCPflags, 0 ) || bit_test( TCPflags, 1 ) || bit_test( TCPflags, 2 ) )
	{
		TCPheaderLength --;
	}


	packetLength-=TCPheaderLength;
	highByte = packetLength >> 8;


#asm
	// increment client SEQ to form our ACK number
	// ( this assembly adds a 16 bit number to a
	// 32 bit number and stores the result in the 32
	// bit number. Based on Microchip App. Note AN617 )																						
	movf	packetLength, 0							
	addwf	headerArray[ CLIENT_SEQ_7to0 ], 1		
													
	movf	highByte, 0								
	btfsc	STATUS, C	
	incfsz	highByte, 0	
	addwf	headerArray[ CLIENT_SEQ_15to8 ], 1	

	clrw
	btfsc	STATUS, C
	movlw	1
	addwf	headerArray[ CLIENT_SEQ_23to16 ], 1

	clrw
	btfsc	STATUS, C
	movlw	1
	addwf	headerArray[ CLIENT_SEQ_31to24 ], 1
#endasm


sendPacket:
	
	replyLength = 40;

	if (data)
	{
		// is file to be sent larger than the assumed MTU?
		if ( fileLength > 536 )
		{
			replyLength = 576;
			dataRemaining = true;
			fileLength -= 536;
			// need to increase the fileStartLow/High bytes after we've sent this chunk

		}
		else
		{
			replyLength += fileLength;
			dataRemaining = false;	// clear the 'more data' flag

			// pad odd length packet
			if ( bit_test( replyLength,0 ) )
			{
				replyLength ++;	
				oddLength = true;									
			}
		}
	}


	startByte = 0;
	endByte = 20;

	putC( END );

	again = false;

	do {
		
		checksumNotDone = true;
		checksumTallyA = 0;
		checksumTallyB = 0;

		do {
		
			if (data)
			{
				i2c_startFunc();
				i2c_write( 0b10100000 );
				i2c_write( fileStartHigh );
				i2c_write( fileStartLow ); 
				i2c_startFunc();
				i2c_write( 0b10100001 );
			}

			counter = startByte;
			sendThisByte = true;

			do {
			
				currentByte = 0;
				
				if ( counter< 52 )
				{
					switch( counter )
					{
					case 0:
						currentByte = 0x45;
						break;
	
					case 50:
						// tcp length for psuedo checksum
						replyLength -= 20;								
					case 2:
						currentByte = replyLength >> 8;
						break;
		
					case 3:
						currentByte = replyLength;
						break;
	
					case 4:
						currentByte = serverIPidentification >> 8;
						break;
	
					case 5:
						currentByte = serverIPidentification;
						break;
	
					case 8:
						currentByte = 64;
						break;
	
					case 9:
						currentByte = PROTOCOL_TCP;
						break;
	
					case 10:
					case 36:
						if ( !checksumNotDone )
						{
							checksumCalc();
						}
						counter+=2;
						continue;

				
					case 40:
						sendThisByte = false;
					case 12:	
						currentByte = SERVER_IP1;
						break;
	
					case 13:
					case 41:
						currentByte = SERVER_IP2;
						break;
	
					case 14:
					case 42:
						currentByte = SERVER_IP3;
						break;
	
					case 15:
					case 43:
						currentByte = SERVER_IP4;
						break;
	
					case 16:
					case 44:
						currentByte = headerArray[ CLIENT_IP1 ];
						break;
				
	
					case 17:
					case 45:
						currentByte = headerArray[ CLIENT_IP2 ];
						break;
	
					case 18:
					case 46:
						currentByte = headerArray[ CLIENT_IP3 ];
						break;
	
					case 19:
					case 47:
						currentByte = headerArray[ CLIENT_IP4 ];
						break;
	
					case 20:
						currentByte = headerArray[ SERVER_PORT_15to8 ];
						break;
	
					case 21:
						currentByte = headerArray[ SERVER_PORT_7to0 ];
						break;
	
					case 22:
						currentByte = headerArray[ CLIENT_PORT_15to8 ];
						break;
	
					case 23:
						currentByte = headerArray[ CLIENT_PORT_7to0 ];
						break;
	
					case 24:
						currentByte = headerArray[ CLIENT_ACK_31to24 ];
						break;

					case 25:
						currentByte = headerArray[ CLIENT_ACK_23to16 ];
						break;

					case 26:
						currentByte = headerArray[ CLIENT_ACK_15to8 ];
						break;

					case 27:
						currentByte = headerArray[ CLIENT_ACK_7to0 ];
						break;

					case 28:
						currentByte = headerArray[ CLIENT_SEQ_31to24 ];
						break;

					case 29:	
						currentByte = headerArray[ CLIENT_SEQ_23to16 ];
						break;

					case 30:
						currentByte = headerArray[ CLIENT_SEQ_15to8 ];
						break;

					case 31:
						currentByte = headerArray[ CLIENT_SEQ_7to0 ];
						break;

					case 32:
					case 34:
						currentByte = 0b01010000;
						break;
	
					case 33:
						currentByte = TCPflags;
						break;
	
					case 49:
						currentByte = PROTOCOL_TCP;
						break;

					case 51:
						currentByte = replyLength;
						replyLength += 20;

					}
				}
				else
				{
					sendThisByte=true;
					currentByte=i2c_read();
				}


				if ( checksumNotDone )
				{
					// odd-numbered byte?
					if ( bit_test( counter,0 ) )		
					{
						checksumTallyB += currentByte;
					}
					else
					{
						checksumTallyA += currentByte;
					}
				}
				else
				{
					// don't send the TCP pseudo header
					if ( sendThisByte )
					{
						sendSLIP( currentByte );		
					}
				}

				counter++;

			} while ( endByte != counter );

			checksumNotDone = !checksumNotDone;

			if (data)
			{
				i2c_read(0);
				i2c_stopFunc();
			}

		} while ( !checksumNotDone );

		again = !again;
				

		// change start and end bytes here!
		startByte = 20;
		endByte = replyLength + 11; // ( == replyLength-1 + 12 )
		if ( !oddLength )
		{
			endByte ++;
		}

	} 
	while( again );


	// pad out with a 0 here?
	if ( oddLength )
	{
		putC( 0 );
		oddLength = false;
	}

	putC( END );

	serverIPidentification ++;


	if ( data )
	{

		
		replyLength -= 40;
		highByte = replyLength >> 8;

		// add the length of data we've sent to our SYN
		#asm
		movf	replyLength, W									
		addwf	headerArray[ CLIENT_ACK_7to0 ], F		

		movf	highByte, W	
		btfsc	STATUS, C	
		incfsz	highByte, W	

		addwf	headerArray[ CLIENT_ACK_15to8 ], F	

		clrw
		btfsc	STATUS, C
		movlw	1
		addwf	headerArray[ CLIENT_ACK_23to16 ], F

		clrw
		btfsc	STATUS, C
		movlw	1
		addwf	headerArray[ CLIENT_ACK_31to24 ], F
		#endasm

		if ( dataRemaining )
		{
			// increase fileStartHigh and fileStartLow to 
			// point to next chunk of data to send from EEPROM
			#asm
			movf	replyLength, W
			addwf	fileStartLow, F
			movf	highByte, W
			btfsc	STATUS, c
			incfsz	highByte, W
			addwf	fileStartHigh, F
			#endasm

		}
		else  // only do these if we've sent the entire file...
		{
			data = FALSE;			
			bit_set( TCPflags, 0 );	// set FIN flag
		}

		goto sendPacket;
	}

	
	goto start;

}